#ifndef CLIENT_H
#define CLIENT_H

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include "pvm3.h"
#include "verbosity_levels.h"

// This file is defined in the superclasses for the server
// (../server directory)
#include "communication_constants.h"

// Subclass off this client to run your client. Your client code will have to
// specify the proper connection to your server and deal with your protocol.
// You will need to use PVM commands to package your messages the PVM hotsheet
// is included in the pvmdocumentation Directory.

// Connection can be made in constructor, or explicitly, to a server of
// server_group running on the specified machine at specified hostname.


class Client {
 private:
    int connected;
    int server_tid;  
    int verbosity_level;
    char server_group[128];
    char server_hostname[128];
 public:
    // Default behavior is to remain unconnected.
    Client();

    // Starts up a client on this machine and, if necessary, a server on
    // a (possibly) remote machine.
    // hostname .......	Name of the (remote) host.
    // group ..........	Server group of server with which to re-establich
    // 			contact.
    // executable .....	Name of the server executable to spawn if attempt
    //			to reconnect fails.
    // verbosity ......	The verbosity level for the duration of the session.
    Client(char *hostname, char *group, char *executable, int verbosity);

    // The default behavior of the servers can handle multiple 
    // disconnection requests (although user may want to override this or
    // free resources associated with a tid or whatever, and this will
    // involve overriding the virtual communicate() call in the server
    // superclasses). If you dont know what I'm talking about I'm just
    // explaining why this behavior is safe in the default server 
    // implementation. If disconnect is going to mean anything special
    // or happen unexpectedly and could screw things up, you have to
    // redesign the server a bit to handle that resource freeing.
    ~Client() {disconnect();}; 

    // Returns server id number.
    int serverTid() {return server_tid;};

    // Predicate controlling printing of error messages.
    int showErrors() {return verbosity_level >= VL_ERRORS;};

    // Predicate controlling tracing.
    int showTrace() {return verbosity_level >= VL_TRACING;};

    // Returns boolean value informing whether the server is connected.
    int isConnected() {return connected;};

    // Returns server group name string.
    char* serverGroup(char* dest = NULL) {
	if (dest) sprintf(dest, "%s", server_group);
	return server_group;
    };

    // Returns hostname for host of server process.
    char* serverHostname(char* dest = NULL) {
	if (dest) sprintf(dest, "%s", server_hostname);
	return server_hostname;
    };
    
    // Disconnects from server by sending a disconnection message.
    // currently, servers dont offer verification for this message.
    // This could be changed. However, if you check the return value on
    // this message, you can be reasonably sure the server got the 
    // message if the call is successful. Returns 0 on success
    // negative on failure
    virtual int disconnect();

    // If you call reconnect with no arguments it reuses the initialization
    // values from constructor or last call to reconnect. Verbosity specifies
    // whether the connection request is documented on stdout...
    virtual int reconnect(char* hostname = NULL, char* group = NULL,
			  int verbosity = 0); 

    // If you have a server_tid, from spawning, say, use this call to hookup.
    virtual int reconnect(int newservertid, int verbosity = 0);
  
    // Tries to spawn the server specified on the hostname
    // this call is not protected, because it might be useful in getting 
    // servers started on multiple machines.
    virtual int spawnserver(char* hostname, char* executable, int verbosity);

    virtual int serverrate(float* rate) {
	pvm_initsend(PvmDataDefault);
	int retval = sendandreceive(REQUEST_MSGTAG_RATE_REPORT);
	if (retval)
	    pvm_unpackf("%d%f",&retval,rate);
	return retval;
    };

    //-------------------------------------------------------------------------
    // SOME NON-ESSENTIAL UTILITIES
    //-------------------------------------------------------------------------
 protected:
    // Wrapper around simple pvm_send.
    // You initialize a buffer and pack it with pvm_initsend() and
    // pvm_pack calls, then call send(). Negative values are system errors.
    virtual int send(int messagetag) {
	return pvm_send(server_tid, messagetag);
    };

    // Wrapper around pvm_recv calls. In general, returns bufid on success and
    // a negative val is failure.  
    // To do a recv that times out, put positive value (microseconds to wait)
    // in timeout parameter. For pure nonblocking you can use timeout of 0,
    // defaults to a negative int which makes for a pure blocking recv
    // in calls that can timeout, 0 means no message is there.  
    // Negative values are errors, otherwise you get bufferid.
    virtual int receive(int messagetag = -1, int timeout = -1)
	{
	    int bufid;
	    if (timeout < 0) {
		bufid = pvm_recv(server_tid, messagetag);
	    } else if (timeout == 0) {
		bufid = pvm_nrecv(server_tid, messagetag);
	    } else {
		timeval tmout;
		div_t timestuff = div(timeout, 1000000);  
		tmout.tv_sec = timestuff.quot;
		tmout.tv_usec = timestuff.rem;
		bufid = pvm_trecv(server_tid, messagetag, &tmout);     
	    };
	    return bufid;  
	};

    // combines the send and receive calls above
    virtual int sendandreceive(int messagetag, int timeout = -1) {
	send(messagetag);
	return (receive(messagetag, timeout));
    };
  
    // Checks to see if a message has arrived
    // retval: 0 means no message is there, egative values are errors,
    // otherwise you get bufferid.
    virtual int probe(int messagetag = -1) {
	return pvm_probe(server_tid, messagetag);
    }; 

};

#endif 
