// $Id: PioneerBridgePrim.cc,v 1.1.1.1 2002/06/26 21:40:30 henrik Exp $
// ***************************************************************************
// *                        A F R O B / P I O N E E R                        *
// *                                                                         *
// *	Module:		PioneerBridgePrim				     *
// *	Purpose:	The C++ core of the PioneerBridge.		     *
// *	Author:		Henrik Nilsson					     *
// *									     *
// ***************************************************************************

// ToDo:
// Not sure about this structure. Eventually, this file could become quite
// large if it contains all APIs and all server start code. However,
// since we need global handles on the client objects to invoke the methods,
// the starting of clients & servers and the APIs are very closely related.
//
// Should adopt some more general conventions/mechanims for tracing
// output (common formatting, similarly named env. variables, etc.).
// The point is that there may be many different bridges, and tracing
// at the level below reactinate can be quite useful. Thus, uniform methods
// for turning tracing and diagnostics on/off would be nice.
//
// It would be nice to simplify this interface by more or less nuking the
// C++ parts and just have a PioneerBridgePrim.c.
//
// It would be nice if verbosity still worked when the communication is
// disabled! Alas, that's a bit messy to do right now. Again, if this
// interface were rewritten by folding in the Client classes, that would
// be easy and natural to do.

// From the Haskell perspective, the PioneerBridge module provides the
// low-level interface to the pioneer. The core of this module (this file)
// is implemented in C++ (relying on the PVM client in the form of the
// class PioneerClient), whereas the Haskell wrapping around this, plus
// any derived Haskell utility functions is implemented in Haskell using FFI.
// To avoid confusion, the externally visible routines in this module have
// the suffix "_prim", whereas the Haskell versions have no suffix.

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "HsFFI.h"
// This ensures that all functions declared in "PioneerDriverCore." (the
// "interface" to this module) will use C *linking* conventions. In particular,
// their names will not be mangled.
extern "C" {
#include "PioneerBridgePrim.h"
}
#include "PioneerClient.h"
#include "tconfig.h"
#include "pvm_gen.h"
#include "pioneer_properties.h"
#include "verbosity_levels.h"

#define NUM_SERVERS 7

// stores information about the running servers - used for restarting
struct ServerTable {
    int tid;			// server task id
    int (*init) (TaskConfig *);	// function to restart server
    TaskConfig *task;
};

static SystemConfig  sys_config;
struct ServerTable   servers[NUM_SERVERS];
static PioneerClient *pioneer = 0;
static bool	     pioneer_enabled = false;


//----------------------------------------------------------------------------
// Pioneer properties
//----------------------------------------------------------------------------

// Note: In the Haskell world, these have type Double and are thus CAFs.
// Hence these functions will only be invoked once.

HsDouble pioneerDiameter_prim(void)
{
    return PIONEER_DIAMETER;
}


HsDouble pioneerMaxAcceleration_prim(void)
{
    return PIONEER_MAX_ACCELERATION;
}


HsDouble pioneerMaxSpeed_prim(void)
{
    return PIONEER_MAX_SPEED;
}


//----------------------------------------------------------------------------
// Starting of the Pioneer server and creation of the client object
//----------------------------------------------------------------------------

// Can also be used for restarting.
static int startPioneerServer(TaskConfig *task)
{
    char *apv, *apd;
    int verb;

    // Figure out verbosity level.
    apv = getenv("AFROB_PIONEER_VERBOSITY");
    if (apv == NULL)
	verb = VL_SILENT;
    else {
	verb = atoi(apv);
	if (verb < VL_SILENT) {
	    verb = VL_SILENT;
	} else if (verb > VL_TRACING) {
	    verb = VL_TRACING;
	}
    }

    // Figure out if the Pioneer should be disabled.
    apd = getenv("AFROB_PIONEER_DISABLE");
    if (apd != NULL && atoi(apd) != 0) {
	pioneer_enabled = false;
	if (verb >= VL_TRACING) {
	    fprintf(stderr, "startPioneerServer: Pioneer disabled!\n");
	}
	return 1;	// Fake tid.
    }

    pioneer = new PioneerClient(task->host_name, task->group_name,
                                task->prog_name, verb);
    if(pioneer->isConnected()) {
	int tid;

	pioneer_enabled = true;
	tid = pioneer->serverTid();
	// Currently we don't do server restart.
	// pvm_notify(PvmTaskExit, DIED_MSG, 1, &tid);
	return tid;
    } else {
	pioneer_enabled = false;
	return 0;
    }
}


//----------------------------------------------------------------------------
// Pioneer server API
//----------------------------------------------------------------------------

HsDouble pioneerServerRate_prim(void)
{
    float rate = 0.0;

    if (pioneer_enabled) {
	assert (pioneer != 0);
	pioneer->serverrate(&rate);
    }
    return (HsDouble) rate;
}


// Odometry

HsBool pioneerResetOdometry_prim(void)
{
    HsBool success = 1;

    if (pioneer_enabled) {
	assert (pioneer != 0);
	success = (HsBool) pioneer->reset_pos();
    }
    return success;
}


HsBool pioneerGetPos_prim(HsPtr xp, HsPtr yp, HsPtr thetap)
{
    HsBool success = 1;
    float x = 0.0, y = 0.0, theta = 0.0;

    if (pioneer_enabled) {
	assert (pioneer != 0);
	success = (HsBool) pioneer->query_pos(&x, &y, &theta);
    }
    // diff_vm returns position in cm and heading in radians.
    *((HsDouble*) xp) = (HsDouble) (x/100);
    *((HsDouble*) yp) = (HsDouble) (y/100);
    *((HsDouble*) thetap) = (HsDouble) theta;
    return success;
}


// Motion

HsBool pioneerSetVelTR_prim(HsDouble v, HsDouble rv)
{
    HsBool success = 1;
    float lwv, rwv;

    // diff_vm expects speed in units of 2 cm/s.
    rwv = (float) (v + (PIONEER_DIAMETER/2)*rv) * 50;
    lwv = (float) (v - (PIONEER_DIAMETER/2)*rv) * 50;
    if (pioneer_enabled) {
	assert (pioneer != 0);
	success = (HsBool) pioneer->diff_vm(lwv, rwv);
    }
    return success;
}


HsBool pioneerSetVelTRGetPos_prim(HsDouble v, HsDouble rv,
                                  HsPtr xp, HsPtr yp, HsPtr thetap)
{
    HsBool success = 1;
    float lwv, rwv, x = 0.0, y = 0.0, theta = 0.0;

    // diff_vm expects speed in units of 2 cm/s.
    rwv = (float) (v + (PIONEER_DIAMETER/2)*rv) * 50;
    lwv = (float) (v - (PIONEER_DIAMETER/2)*rv) * 50;
    if (pioneer_enabled) {
	assert (pioneer != 0);
	success = (HsBool) pioneer->diff_vm(lwv, rwv, &x, &y, &theta);
    }
    // diff_vm returns position in cm and heading in radians.
    *((HsDouble*) xp) = (HsDouble) (x/100);
    *((HsDouble*) yp) = (HsDouble) (y/100);
    *((HsDouble*) thetap) = (HsDouble) theta;
    return success;
}


HsBool pioneerSetVelDiff_prim(HsDouble lwv, HsDouble rwv)
{
    HsBool success = 1;

    if (pioneer_enabled) {
	assert (pioneer != 0);
	// diff_vm expects speed in units of 2 cm/s.
	success = (HsBool) pioneer->diff_vm((float)(lwv*50), (float)(rwv*50));
    }
    return success;
}


HsBool pioneerSetVelDiffGetPos_prim(HsDouble lwv, HsDouble rwv,
                                    HsPtr xp, HsPtr yp, HsPtr thetap)
{
    HsBool success = 1;
    float x = 0.0, y = 0.0, theta = 0.0;

    if (pioneer_enabled) {
	assert (pioneer != 0);
	// diff_vm expects speed in units of 2 cm/s.
	success = (HsBool) pioneer->diff_vm((float) (lwv*50), (float) (rwv*50),
					    &x, &y, &theta);
    }
    // diff_vm returns position in cm and heading in radians.
    *((HsDouble*) xp) = (HsDouble) (x/100);
    *((HsDouble*) yp) = (HsDouble) (y/100);
    *((HsDouble*) thetap) = (HsDouble) theta;
    return success;
}


HsBool pioneerStop_prim(void)
{
    HsBool success = 1;

    if (pioneer_enabled) {
	assert (pioneer != 0);
	success = (HsBool) pioneer->st();
    }
    return success;
}


HsBool pioneerStopGetPos_prim(HsPtr xp, HsPtr yp, HsPtr thetap)
{
    HsBool success = 1;
    float x = 0.0, y = 0.0, theta = 0.0;

    if (pioneer_enabled) {
	assert (pioneer != 0);
	success = (HsBool) pioneer->st(&x, &y, &theta);
    }
    // diff_vm returns position in cm and heading in radians.
    *((HsDouble*) xp) = (HsDouble) (x/100);
    *((HsDouble*) yp) = (HsDouble) (y/100);
    *((HsDouble*) thetap) = (HsDouble) theta;
    return success;
}


//----------------------------------------------------------------------------
// Starts all PVM servers according to the configuration file
//----------------------------------------------------------------------------

// This one should return an exit code indicating success of failure. Bool?

HsBool startServers_prim(void)
{
    TaskConfig *task;
    HsBool     success;
    int        i;

    // Reads the system.cfg file to get the configuration.
    if(get_system_config(SYS_FILE_NAME, &sys_config) != 0) {
	return (HsBool) 0;
    }

    // Initialize table of restarting functions. Not currently used.  
    // servers[0].init = &init_vision;
    // servers[1].init = &init_pan;
    servers[2].init = &startPioneerServer;
    // servers[3].init = &restart_laser;
    // servers[4].init = &restart_selection;
    // servers[5].init = &restart_mapping;
    // servers[6].init = &restart_matlab;

    success = 1;
    for(i=0, task=&(sys_config.task[0]); i<sys_config.num_tasks; i++, task++) {
	switch(task->type) {
	case TSERV_TASK:
	    // servers[0].tid = init_vision(task);
	    // servers[0].task = task;
	    fprintf(stderr,
		    "Currently no vision: "
		    "ignoring request to start Vision server.\n");
	    break;
	case PSERV_TASK:
	    // servers[1].tid = init_pan(task);
	    // servers[1].task = task;
	    fprintf(stderr,
		    "Currently no panning: "
		    "ignoring request to start Pan server.\n");
	    break;
	case SSERV_TASK:
	    servers[2].tid = startPioneerServer(task);
	    servers[2].task = task;
	    if (servers[2].tid == 0) success = 0;
	    break;
	case LASER_TASK:
	    // servers[3].tid = init_laser(task);
	    // servers[3].task = task;
	    fprintf(stderr,
		    "Currently no laser: "
		    "ignoring request to start Laser server.\n");
	    break;
	case COLORSEL_TASK:
	    // servers[4].tid = init_selection(task);
	    // servers[4].task = task;
	    fprintf(stderr,
		    "Currently no color selection: "
		    "ignoring request to start Color server.\n");
	    break;
	case MAP_TASK:
	    // servers[5].tid = init_mapping(task);
	    // servers[5].task = task;
	    fprintf(stderr,
		    "Currently no mapping: "
		    "ignoring request to start Map server.\n");
	    break;
	case TRACE_TASK:
	    // init_tracer(task);
	    fprintf(stderr,
		    "Currently no tracing: "
		    "ignoring request to start Tracer server.\n");
	    break;
	case MATLAB_TASK:
	    // servers[6].tid = init_matlab(task);
	    // servers[6].task = task;	
	    fprintf(stderr,
		    "Currently no MatLab support: "
		    "ignoring request to start MatLab server.\n");
	    break;
	default:
	    break;
	}
    }
    return success;
}

// Might want something like this at some point? (E.g. for server restart.)
// But cannot rely on Qt in our case.
//
// Periodic Timer event called by QT
//
// void commander::timerEvent( QTimerEvent *e )
// {
//   int status;
//   if( e->timerId() ==  battery_timer)
//     timerBattery();      // update battery status
//   else {
//     timerServerChk();    // check if any server died
//     update_perfdisp();   // update performance display, if active
//   }
// }
// 
// // Checks if some server has died, and attempts to restart it
// void timerServerChk(void)
// {
//   int msgtag, tid, bufid, bytes;
//   int task_id;
// 
//   // receive message about server death, if any
//   if(!(bufid = pvm_nrecv(-1,DIED_MSG))) return;
//
//   // Note: pufid could be an error code too. Should check that.
//
//   pvm_bufinfo(bufid,&bytes,&msgtag,&tid);
// 
//   // if a server has died
//   if (msgtag == DIED_MSG) {
//     pvm_upkint(&task_id,1,1);	// receive tid of dead server
// 
//     cout << "Server with tid " << task_id << " died" << endl;
// 
//     // check the tids of all servers and restart the dead one
//     for (int i=0; i<NUM_SERVERS; i++)
// 	 if (task_id == servers[i].tid)
// 	 {
// 	   struct timeval time={0,200};
// 	   
// 	   // clear buffers
// 	   while(pvm_trecv(task_id,-1,&time));
// 	   servers[i].tid = servers[i].init(servers[i].task);
// 	 }
//   }
// }
