// There should probably be a class RobotClient or DiffDriveRobotClient
// for capturing common aspects of Scouts/Pioneers/similar, and then
// subclasses ScoutClient and PioneerClient of that.

#ifndef PIONEERCLIENT_H
#define PIONEERCLIENT_H

#include <iostream.h>
#include "pvm3.h"
#include "Client.h" 
#include "communication_constants.h"
#include "ss_communication_constants.h"

class PioneerClient : public Client {
 public:
    PioneerClient();
    PioneerClient(char *hostname, char *group, char *executable, int verbosity);
    ~PioneerClient();

    /* ********************** MOTION COMMANDS ********************/
    // There ought to be a more clearly defined set of operations
    // implementable both on the Pinoeer and the Scout. In particular,
    // one should agree on units. Preferably m, m/s, rad, rad/s, etc.
    // For now, rely mainly on diff_vm.

    // Velocity Mode sets translational and rotational velocities
    // (the server converts this into left and right wheel velocities).
    // The units seem to be "Pioneer units": the floats are simply typecasted
    // to (signed) char on the server side, and the right and left wheel
    // velocities are computed as a simple sum and difference of the
    // translational and rotational velocities without taking the robot
    // geometry into consideration.
    // Optional return value is position data (in cm and radians).
    int vm(float v, float rv,
	   float *xp = NULL, float *yp = NULL, float *thetap = NULL);

    // Velocity mode that gives you access to wheels separately.
    // Use in preference to vm.
    // lwv ....	Left Wheel Velocity [2 cm/s].
    // rwv ....	Right Wheel Velocity [2 cm/s].
    // x ......	X position [cm].
    // y ......	Y position [cm].
    // theta .. Heading [rad].
    int diff_vm(float lwv, float rwv,
		float *xp = NULL, float *yp = NULL, float *thetap = NULL);

    // Limps motors (seemingly non-functional).
    int lp(float *xp = NULL, float *yp = NULL, float *thetap = NULL);

    // Stops robot motion.
    int st(float *xp = NULL, float *yp = NULL, float *thetap = NULL);
    /* *********************************************************/


    /* ******************* POSITION ****************************/
    // Set odometry. The Pioneer will IGNORE the arguments and reset to 0,0,0.
    int set_pos(float x, float y, float theta);

    // Use in preference to set_pos.
    int reset_pos() {return set_pos(0, 0, 0);};

    // Get position according to odometry.
    int query_pos(float *xp = NULL, float *yp = NULL, float *thetap = NULL);

    /* *********************************************************/


    /* ***************** MOTION CONTROL *************************/
    // used with one argument, accelerations are set to same value
    // int set_wheel_acceleration(int lefta, int righta = -1);
    // int reset_wheel_acceleration() {return set_wheel_acceleration(75);};

    // used to set wheel velocities, used in pr commands only
    // int set_wheel_velocity(int leftv, int rightv = -1);
    // int reset_wheel_velocity() {return set_wheel_velocity(50);};
    /* *********************************************************/  


    /* *************** SONAR COMMANDS *************************/
    // turn on front 13 sonars
    // int turn_front_sonar_on(int rate =15);
    // turn on front 5 sonars
    // int turn_frontfive_sonar_on(int rate =15);
    // turn on all 16 sonars
    // int turn_all_sonar_on(int rate = 15);  
    // turn on sonars with specified configuration values
    // int  specify_sonar_order(int* sonar_firing_order, int rate = 15);
    // turn sonars (and sonar noise) off
    // int turn_sonar_off();

    // gets all 16 sonar values, regardless of how they are configured
    // returns value in memory pointed to by sonar_vector
    // int get_sonar(float* sonar_vector);
    /* *********************************************************/  


    /*  ************* BUMPER, STATUS COMMANDS ******************/  
    // get bumper vector
    // returns value in memory pointed to by bumper_vector
    // int get_bumpers(int* bumper_vector);
  
#define RIGHT_WHEEL_IN_MOTION 0
#define LEFT_WHEEL_IN_MOTION 1
#define BATTERY_0 2
#define BATTERY_1 3
#define PLUGGED_INTO_AC 4
#define CHARGING 5
#define EMERGENCY_STOP 6
    // get status of bit in state vector for the quantities defined above 
    // returns value in memory pointed to by status argument
    int get_status(int what, int* status);
  
#define LOW_BATTERY 0
#define MED_BATTERY 1
#define HIGH_BATTERY 2 
#define RESERVED 3
    // return info on battery level for robot, 
    // returns value in memory pointed to by status argument
    int get_battery_status(int* status);
    /* *********************************************************/  
     
};

#endif 
